/* daemonitor/src/daemonitor/200_log.c
 *
 *  (c)2007, Laurence Withers, <l@lwithers.me.uk>.
 *  Released under the GNU GPLv3. See file COPYING or
 *  http://www.gnu.org/copyleft/gpl.html for details.
*/



/* LOGGING ROUTINES *******************************************************************************
 *
 *  This file provides a set of routines for performing logging. Messages are logged whenever
 *  actions are taken, such as setting up or shutting down. Messages can be logged to syslog,
 *  stdout, stderr or a file.
 *
 *  Errors that occur before logging is set up (e.g. invalid commandline options, or opening a log
 *  file) will be reported to both stderr and syslog (see LOG_ANYWHERE()).
 */



/* PUBLIC API ************************************************************************************/



/* enum log_destination_t
 *  The possible log destinations. These are passed to log_destination_set(), below.
 */
enum log_destination_t {
    log_destination_stdout,
    log_destination_stderr,
    log_destination_syslog,
    log_destination_file
};



/* log_destination_set()
 *  Chooses a specific log destination. If the destination is a file, log_destination_set_file()
 *  must also be called.
 */
void log_destination_set(enum log_destination_t dest);



/* log_destination_set_file()
 *  Sets the output file for logging. Returns 0 on success, or -1 on error (i.e. if the file cannot
 *  be opened for writing), with errno set to whatever error caused the open to fail.
 */
int log_destination_set_file(const char* filename);



/* LOG()
 *  The standard logging function. This sends the logging output to whatever destination has been
 *  chosen. It can only be called after logging has been set up; for other errors, LOG_ANYWHERE
 *  must be called.
 */
#define LOG(level, fmt, ...) \
    log_func(level, "%s:%d: " fmt, __FUNCTION__, __LINE__ , ## __VA_ARGS__ )



/* LOG_ANYWHERE()
 *  Try to log a problem, anywhere. Try stderr and syslog.
 */
#define LOG_ANYWHERE(fmt, ...) \
    log_anywhere("%s:%d: " fmt, __FUNCTION__, __LINE__ , ## __VA_ARGS__ )



/* IMPLEMENTATION ********************************************************************************/



/* log_anywhere()
 *  Logs to both stderr and syslog. Used by LOG_ANYWHERE() macro.
 */
void log_anywhere(const char* fmt, ...) __attribute__((format(printf,1,2),nonnull));
void log_anywhere(const char* fmt, ...)
{
    va_list va, va2;

    /* get two copies of argument list */
    va_start(va, fmt);
    va_copy(va2, va);

    /* print to stderr */
    vdprintf(STDERR_FILENO, fmt, va);
    safe_write_fixed(STDERR_FILENO, "\n", 1);
    va_end(va);

    /* send to syslog */
    vsyslog(LOG_CRIT, fmt, va2);
    va_end(va2);
}



/* log_func
 *  This is a pointer to the actual logging function. The pointer is set by log_destination_set().
 */
void (*log_func)(int level, const char* fmt, ...) __attribute__((nonnull,format(printf,2,3)));



/* log_func_fd, log_func_file()
 *  Logs to a file descriptor.
 */
int log_func_fd;
void log_func_file(int level, const char* fmt, ...)
{
    va_list va;
    int ret, saved_errno;
    const char* level_str;
    char timebuf[40];
    time_t t;

    /* save errno, because dprintf will clobber it */
    saved_errno = errno;

    /* choose appropriate line description */
    switch(level) {
    case LOG_EMERG:     level_str = "EMERG"; break;
    case LOG_ALERT:     level_str = "ALERT"; break;
    case LOG_CRIT:      level_str = "CRIT "; break;
    case LOG_ERR:       level_str = "ERROR"; break;
    case LOG_WARNING:   level_str = "WARN "; break;
    case LOG_NOTICE:    level_str = "NOTE "; break;
    case LOG_INFO:      level_str = "INFO "; break;
    case LOG_DEBUG:     level_str = "DEBUG"; break;
    default:            level_str = "?BUG?"; break;
    }

    /* construct timestamp */
    time(&t);
    strftime(timebuf, sizeof(timebuf), "%FT%H:%M:%S", gmtime(&t));

    /* print to file descriptor */
    va_start(va, fmt);
    ret = dprintf(log_func_fd, "%s %s ", timebuf, level_str);
    errno = saved_errno;
    if(ret != -1) ret = vdprintf(log_func_fd, fmt, va);
    if(ret != -1 && safe_write_fixed(log_func_fd, "\n", 1) != 1) ret = -1;
    va_end(va);

    /* report errors */
    if(ret < 0) {
        log_anywhere("error writing to log file descriptor %d (%m)", log_func_fd);
    }
}



/* log_func_syslog()
 *  Logs to syslog.
 */
void log_func_syslog(int level, const char* fmt, ...)
{
    va_list va;

    /* log to syslog */
    va_start(va, fmt);
    vsyslog(level, fmt, va);
    va_end(va);
}



/* log_destination_set()
 *  Set `log_func' function pointer.
 */
void log_destination_set(enum log_destination_t dest)
{
    switch(dest) {
    case log_destination_stdout:
        log_func_fd = STDOUT_FILENO;
        log_func = log_func_file;
        break;

    case log_destination_stderr:
        log_func_fd = STDERR_FILENO;
        log_func = log_func_file;
        break;

    case log_destination_syslog:
        log_func = log_func_syslog;
        break;

    case log_destination_file:
        /* log_func_fd() must be set by log_destination_set_file() */
        log_func = log_func_file;
        break;
    }
}



/* log_destination_set_file()
 *  Open `filename' for writing, return 0 on success and -1 on error. sets `log_func_fd'.
 */
int log_destination_set_file(const char* filename) __attribute__((nonnull,warn_unused_result));
int log_destination_set_file(const char* filename)
{
    log_func_fd = open(filename, O_WRONLY | O_CREAT | O_APPEND, 0666);
    if(log_func_fd == -1) return -1;
    return 0;
}



/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
