/* daemonitor/src/daemonitor/600_run_program.c
 *
 *  (c)2007, Laurence Withers, <l@lwithers.me.uk>.
 *  Released under the GNU GPLv3. See file COPYING or
 *  http://www.gnu.org/copyleft/gpl.html for details.
*/



/* PROGRAM EXECUTION AND MONITORING ROUTINES ******************************************************
 *
 *  This file provides the run_program() routine, which actually runs and waits on the child
 *  process. `program_pid' is set to the PID of the child, which allows it to be sent signals.
 */



/* PUBLIC API ************************************************************************************/



/* run_program()
 *  Runs the program with the specified arguments. Logs the start and end of the process, including
 *  exit status. Waits on the process to exit. A single call to this function is a single run. If
 *  the program cannot be run, for whatever reason, that is logged and the function returns.
 */
void run_program(const char* exe_path, char** argv);



/* program_pid
 *  Set to (pid_t)0 if there is no currently active program, otherwise set to the PID of the running
 *  child process. This is intended for use in signal handlers.
 */
extern pid_t program_pid;



/* sigterm_received
 *  Set by a signal handler if SIGTERM is received while program_pid is set, this flag causes
 *  run_program() to shut the child process down.
 */
volatile int sigterm_received = 0;



/* close_file_descriptors()
 *  Closes all file descriptors but stdin, stdout and stderr. Should be called whether we are
 *  daemonising or not.
 */
void close_file_descriptors(void);



/* IMPLEMENTATION ********************************************************************************/



/* crash()
 *  Crash the program. This is better than just calling exit() because it doesn't clean up. We
 *  should leave our PID file lying around to look like we crashed. This function is paranoid.
 */
void crash(void) __attribute__((noreturn));
void crash(void)
{
    /* raise SIGKILL; failing that, call abort(); failing that, _exit() */
    raise(SIGKILL); /* quick + dirty + unblockable crash */
    abort(); /* unblocks and raises SIGABRT */
    _exit(127);
}



/* program_pid, program_pid_lock()
 *  Used to record the PID of the child process for signal handlers. 0 means there is no child
 *  process running. The lock function actually blocks/unblocks those signals whose handlers would
 *  examine program_pid; they can be used to avoid races.
 */
pid_t program_pid = 0;

void program_pid_lock(int lock)
{
    sigset_t ss;

    sigemptyset(&ss);
    sigaddset(&ss, SIGTERM);

    sigprocmask(lock ? SIG_BLOCK : SIG_UNBLOCK, &ss, 0);
}



/* run_program()
 *  fork() then wait(), logging the results of each step.
 */
void run_program(const char* exe_path, char** argv)
{
    pid_t pid;
    siginfo_t si;
    sigset_t ss;
    int i;

    program_pid_lock(1);

    pid = fork();

    /* deal with errors */
    if(pid == (pid_t)-1) {
        program_pid_lock(0);
        LOG(LOG_ERR, "Could not create new child process (%m).");
        return;
    }

    /* child -- execute new process */
    if(pid == 0) {
        /* clear the child's signal mask */
        sigemptyset(&ss);
        sigprocmask(SIG_SETMASK, &ss, 0);

        execve(exe_path, argv, child_environ);
        LOG(LOG_ERR, "Could not execute `%s' (%m).", exe_path);
        crash();
    }

    /* parent -- wait(2) on child */
    program_pid = pid;
    program_pid_lock(0);
    LOG(LOG_NOTICE, "Created child process with PID %lu.", (unsigned long)pid);

  wait_again:
    /* wait on child */
    si.si_pid = 0;
    if(waitid(P_PID, pid, &si, WEXITED) == -1) {
        switch(errno) {
        case EINTR:
            if(sigterm_received) goto sigterm;
            break; /* signal processed */

        default:
            LOG(LOG_ALERT, "waitid() returned error `%m' for PID %lu -- should not be possible.",
                (unsigned long)pid);
            crash();
        }
        goto wait_again; /* go again */
    }

    /* check that child did in fact exit */
    if(si.si_pid != pid) {
        LOG(LOG_ALERT, "waitid() returned results for PID %lu, but we were expecting %lu.",
            (unsigned long)si.si_pid, (unsigned long)pid);
        crash();
    }

    /* report how the child exited */
    switch(si.si_code) {
    case CLD_EXITED:
        LOG(LOG_NOTICE, "Child process %lu exited with status %d.",
            (unsigned long)pid, si.si_status);
        break;

    case CLD_KILLED:
        LOG(LOG_NOTICE, "Child process %lu killed by signal %d (%s).",
            (unsigned long)pid, si.si_status, strsignal(si.si_status));
        break;

    default:
        LOG(LOG_WARNING, "Received si_code 0x%X from waitid(); ignoring it.", si.si_code);
        goto wait_again;
    }

    /* deal with SIGTERM, avoiding race */
    program_pid_lock(1);
    program_pid = 0;
    program_pid_lock(0);
    if(sigterm_received) exit(0);
    return;

  sigterm:
    /* SIGTERM was received, send SIGTERM to the child */
    LOG(LOG_NOTICE, "Termination signal received, passing it to child.");
    kill(program_pid, SIGTERM);

    /* wait for up to 3 seconds for it to exit */
    si.si_pid = 0;
    for(i = 0; i < 30; ++i) {
        waitid(P_PID, pid, &si, WNOHANG | WEXITED);
        if(si.si_pid) {
            LOG(LOG_NOTICE, "Child has exited. Shutting down.");
            exit(0);
        }

        safe_sleep_fixed(0, 100 * 1000 * 1000);
    }

    /* now send it a SIGKILL */
    LOG(LOG_WARNING, "Child process did not exit. Sending it a SIGKILL.");
    errno = 0;
    kill(program_pid, SIGKILL);
    LOG(LOG_WARNING, "Result: %m.");
    exit(0);
}



/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
