/* daemonitor/src/daemonitor/999_main.c
 *
 *  (c)2007, Laurence Withers, <l@lwithers.me.uk>.
 *  Released under the GNU GPLv3. See file COPYING or
 *  http://www.gnu.org/copyleft/gpl.html for details.
*/



/* DEFAULT_RESTART_INTERVAL
 *  The default argument to --restart-interval, in seconds.
 */
#define DEFAULT_RESTART_INTERVAL 10
#define DEFAULT_RESTART_INTERVAL_S "10"



/* usage()
 *  Display the usage info.
 */
void usage(void)
{
    fputs("Usage:\n\n"
        "    daemonitor [options] -- /path/to/program [args]\n\n"
        "Valid options:\n"
        " -h, --help                Display this screen.\n"
        " -V, --version             Display version number.\n"
        " -d, --daemon              Daemonise.\n"
        " -l, --log <dest>          Log file destination, may be one of:\n"
        "                               SYSLOG:name - log to syslog\n"
        "                               STDOUT - log to stdout\n"
        "                               STDERR - log to stderr (default)\n"
        "                               file path - log to file\n"
        " -p, --pidfile <path>      Write PID to file.\n"
        " -R, --restart-interval <i>   Sets the length of time between the process\n"
        "                              exiting and being restarted, in seconds (default\n"
        "                              " DEFAULT_RESTART_INTERVAL_S "). May be 0 to restart immediately.\n"
        " -E, --environment <path>  Name of file containing environment variables for\n"
        "                           child process, one NAME=VALUE entry per line.\n"
    "", stderr);
}



/* options[]
 *  The array of long commandline options for getopt(3).
 */
struct option options[] = {
    { "help", no_argument, 0, 'h' },
    { "version", no_argument, 0, 'V' },
    { "daemon", no_argument, 0, 'd' },
    { "log", required_argument, 0, 'l' },
    { "pidfile", required_argument, 0, 'p' },
    { "restart-interval", required_argument, 0, 'R' },
    { "environment", required_argument, 0, 'E' },
    { 0, 0, 0, 0 }
};



/* optstr
 *  The list of short commandline options for getopt(3). Sorted alphabetically, lowercase first.
 */
const char* optstr = "dhl:p:E:R:V";



/* main()
 *  The program driver. Parses the commandline options, sets up the environment, and starts the
 *  daemon monitoring process.
 */
int main(int argc, char* argv[])
{
    char* endp;
    const char* log_argument = 0, * pidfile_argument = 0, * restart_argument = 0;
    int daemon = 0, restart_interval = DEFAULT_RESTART_INTERVAL;

    /* safety */
    close_file_descriptors(); /* leaves stdout etc. open */

    /* parse commandline arguments */
    while(1) {
        switch(getopt_long(argc, argv, optstr, options, 0)) {
        case -1: goto opts_done;
        case '?': LOG_ANYWHERE("Invalid commandline options."); return 1;
        case 'h': usage(); return 1;
        case 'V': fputs("daemonitor " VERSION "\n", stderr); return 1;
        case 'd': ++daemon; break;
        case 'l': log_argument = optarg; break;
        case 'p': pidfile_argument = optarg; break;
        case 'R': restart_argument = optarg; break;
        case 'E': set_environment_file(optarg); break;
        }
    }
  opts_done:
    if(restart_argument) {
        restart_interval = strtol(restart_argument, &endp, 0);
        if(*endp || restart_interval < 0) {
            LOG_ANYWHERE("Invalid --restart-interval option `%s'. Must be an integer >= 0.",
                restart_argument);
            return 1;
        }
    }

    if(optind >= argc) {
        LOG_ANYWHERE("Program not specified.");
        return 1;
    }

    /* set up logging */
    if(!log_argument || !strcmp(log_argument, "STDERR")) {
        log_destination_set(log_destination_stderr);
    } else if(!strcmp(log_argument, "STDOUT")) {
        log_destination_set(log_destination_stdout);
    } else if(!strncmp(log_argument, "SYSLOG:", 7)) {
        /* force syslog to be opened now, so it doesn't get opened halfway through daemonise() */
        openlog(log_argument + 7, LOG_NDELAY, LOG_DAEMON);
        log_destination_set(log_destination_syslog);
    } else {
        if(log_destination_set_file(log_argument)) {
            LOG_ANYWHERE("Error logging to file `%s' (%m).", log_argument);
            return 1;
        }
        log_destination_set(log_destination_file);
    }

    /* set up environment */
    if(load_child_environ()) return 1;

    /* daemonise? */
    if(daemon) daemonise();

    /* write a PID file */
    if(pidfile_argument) {
        pidfile_set_filename(pidfile_argument);
        /* keep retrying until we're done */
        while(pidfile_write()) {
            safe_sleep_fixed(5, 0);
        }
    }

    /* set up signals */
    signal_setup();

    /* run process */
    while(1) {
        run_program(argv[optind] /* path */, argv + optind /* argv[], null terminated */);

        if(restart_interval) {
            LOG(LOG_NOTICE, "Sleeping for %d seconds.", restart_interval);
            safe_sleep_fixed(restart_interval, 0);
        }
    }

    return 0;
}



/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
